#!/usr/bin/python3

# Copyright 2021 Offensive Security
# SPDX-license-identifier: GPL-3.0-only

import logging
import os
import sys
import termios
from contextlib import contextmanager

from snack import *

from .toolkit import (
    KaliCheckboxBar,
    KaliConfirmWindow,
    KaliListboxWindow,
    KaliInfoWindow,
    KaliListbox,
    KaliRadioBar,
    KaliWindow,
    KaliSimpleCheckboxTree,
)
from .utils import *

LOG_FILE = "kali-tweaks-log.txt"
LOG_LEVEL = logging.DEBUG

if "DEBUG" in os.environ:
    # Can't use stdout while the UI is on, so we send logs to a file instead
    logging.basicConfig(filename=LOG_FILE, filemode="w", level=LOG_LEVEL)

logger = logging.getLogger("kali-tweaks")


### HELPERS


def get_metapackages(search):
    """Return an array of (pkg, description, installed)"""
    out = []

    # Run the command
    res = run(search)

    # If exit code wasn't expected, quit
    if res.returncode != 0:
        return out

    # For every result
    for line in res.stdout.splitlines():
        try:
            # Example: kali-linux-default - Kali Linux default system
            pkg, desc = line.split(" - ", maxsplit=1)
        except ValueError:
            continue
        # Check to see if installed
        installed = is_pkg_installed(pkg)
        # Record
        out += [(pkg, desc, installed)]

    return out


def get_metapackage_items():
    items = []

    # List the Kali-tools
    tools = get_metapackages(
        "apt-cache search --names-only ^kali-tools | grep -v ' menu$'"
    )
    for pkg, desc, installed in tools:
        short_pkg = pkg.removeprefix("kali-tools-")
        short_desc = desc.removeprefix("Kali Linux ")
        items += [(short_pkg, short_desc, installed)]

    # List the Kali-linux metapackages
    tools = get_metapackages(
        "apt-cache search --names-only ^kali-linux | grep -vE '\-(arm|nethunter|core|top10) '"
    )
    for pkg, desc, installed in tools:
        items += [(pkg, desc, installed)]

    return items


#### UI HELPERS


@contextmanager
def suspend(screen, ending_message=None, wait_enter=False):
    """
    Suspend the UI while running a command. This is useful to show progress
    for long operations (package installation), or when user interaction is
    required (for example, entering a password).
    """
    screen.suspend()
    try:
        yield
    finally:
        if ending_message:
            print()
            print("┏━(\033[1;31mMessage from Kali developers\033[00m)")
            for line in ending_message.splitlines():
                print("┃ " + line)
            print("┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━")
            print()
            wait_enter = True
        if wait_enter:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)
            input(bold("> Press Enter to continue..."))
        screen.resume()


def say_reload_terminal(screen, first_line):
    msg = f"{first_line}\n"
    msg += "Please close and restart all your terminal instances"
    msg += " for the change to take effect."
    KaliInfoWindow(screen, "info", msg)


def say_log_out(screen, first_line):
    msg = f"{first_line}\n"
    msg += "Please log out and log back in"
    msg += " for the change to take effect."
    KaliInfoWindow(screen, "info", msg)


#### CLASSES


class KaliShellDefaultLoginContent(Grid):
    def __init__(self, config):
        # XXX Vertical padding should be automated somehow
        super().__init__(1, 2)

        # Get the values we care about
        shell = config.get("shell")

        # Build the form with all the settings
        heading = "Default Login Shell to use:"
        items = [
            ("Bash", "Bourne Again SHell", "bash"),
            ("ZSH", "Z SHell", "zsh"),
        ]
        radioboxes = KaliRadioBar(heading, items, shell)
        self.setField(radioboxes, 0, 0, growx=1)

        vertical_padding = Textbox(1, 9, "")
        self.setField(vertical_padding, 0, 1, growx=1)

        # Save a few things for later
        self.orig_config = config
        self.radioboxes = radioboxes

    def getSelection(self):
        selected = self.radioboxes.getSelection()
        config = {}
        config["shell"] = selected
        return config

    def getChanges(self):
        changes = {}
        orig_config = self.orig_config
        new_config = self.getSelection()

        for key in new_config:
            if new_config[key] != orig_config[key]:
                changes[key] = new_config[key]

        logger.debug("orig_config: %s", orig_config)
        logger.debug("new_config : %s", new_config)
        logger.debug("changes    : %s", changes)

        return changes


class KaliShellPromptContent(Grid):
    def __init__(self, config):
        # XXX Vertical padding should be automated somehow
        super().__init__(1, 3)

        # Get the values we care about
        newline = config.get("newline", True)
        prompt = config.get("prompt", "twoline")

        # Build the form with all the settings
        heading = "Prompt Style:"
        items = [
            ("Two Lines", "Dual blue line prompt", "twoline"),
            ("One Line", "Single blue line prompt", "oneline"),
            ("BackTrack", "Legacy BackTrack red prompt", "backtrack"),
        ]
        radioboxes = KaliRadioBar(heading, items, prompt)
        self.setField(radioboxes, 0, 0, growx=1)

        heading = "Prompt Settings:"
        items = [
            ("Newline", "Add a new line between output and prompt", newline),
        ]
        checkboxes = KaliCheckboxBar(heading, items)
        self.setField(checkboxes, 0, 1, growx=1)

        vertical_padding = Textbox(1, 6, "")
        self.setField(vertical_padding, 0, 2, growx=1)

        # Save a few things for later
        self.orig_config = config
        self.radioboxes = radioboxes
        self.checkboxes = checkboxes

    def getSelection(self):
        config = {}
        config["prompt"] = self.radioboxes.getSelection()
        config["newline"] = self.checkboxes.getSelection()[0]
        return config

    def getChanges(self):
        changes = {}
        orig_config = self.orig_config
        new_config = self.getSelection()

        for key in new_config:
            # WARNING! No guarantee that orig config had all keys.
            # At this moment, keys might be missing.
            if not key in orig_config:
                changes[key] = new_config[key]
            elif new_config[key] != orig_config[key]:
                changes[key] = new_config[key]

        logger.debug("orig_config: %s", orig_config)
        logger.debug("new_config : %s", new_config)
        logger.debug("changes    : %s", changes)

        return changes


class KaliNetworkRepoContent(Grid):
    def __init__(self, config):
        # XXX Vertical padding should be automated somehow
        super().__init__(1, 2)

        # Get the values we care about
        bleeding_edge = config["bleeding-edge"]
        experimental = config["experimental"]

        # Build the form with all the settings
        heading = "Additional Kali repositories:"
        items = [
            (
                "bleeding-edge",
                "Automatically packaged and potentially unstable",
                bleeding_edge,
            ),
            (
                "experimental",
                "Staging area for work-in-progress packages",
                experimental,
            ),
        ]
        checkboxes = KaliCheckboxBar(heading, items)
        self.setField(checkboxes, 0, 0, growx=1)

        vertical_padding = Textbox(1, 9, "")
        self.setField(vertical_padding, 0, 1, growx=1)

        # Save a few things for later
        self.orig_config = config
        self.checkboxes = checkboxes

    def getSelection(self):
        selected = self.checkboxes.getSelection()
        config = {}
        config["bleeding-edge"] = selected[0]
        config["experimental"] = selected[1]
        return config

    def getChanges(self):
        changes = {}
        orig_config = self.orig_config
        new_config = self.getSelection()

        for key in new_config:
            if new_config[key] != orig_config[key]:
                changes[key] = new_config[key]

        logger.debug("orig_config: %s", orig_config)
        logger.debug("new_config : %s", new_config)
        logger.debug("changes    : %s", changes)

        return changes


#### SECONDARY SCREENS


def do_shell_change_default_login(screen):
    title = "Default Login Shell"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Get effective and configured login shells, make sure there's no change pending
    configured_shell = get_configured_login_shell()
    effective_shell = get_effective_login_shell()
    if configured_shell != effective_shell:
        msg = f"Your shell is configured to be {configured_shell},"
        msg += f" however at the moment it is effectively {effective_shell}.\n"
        msg += "It seems that you have changes pending,"
        msg += " please log out and log back in for these changes to take effect."
        KaliInfoWindow(screen, "warning", msg)
        return

    # Prepare a config
    config = {}
    config["shell"] = effective_shell

    # Show the window with all the settings
    content = KaliShellDefaultLoginContent(config)
    w = KaliWindow(screen, title, content, buttons)
    btn = w.run_once()
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    old_shell = config.get("shell", "unknown")
    new_shell = changes["shell"]
    msg = ""
    msg += f"Your current login shell is {old_shell}.\n"
    msg += f"Changing it to {new_shell}.\n"
    msg += "Proceed?"

    btn = KaliConfirmWindow(screen, msg, title=title)
    if btn != "ok":
        return

    # Do the actual job
    try:
        with suspend(screen):
            change_login_shell(new_shell)
    except RuntimeError as e:
        KaliInfoWindow(screen, "error", str(e))
        return

    # One last notification
    say_log_out(screen, "Your login shell was changed!")


def do_shell_reset(screen):
    configs = get_shell_config_files()

    title = "Reset Shell Config"
    msg = "The following config files will be replaced by Kali's default:\n"
    for default, current in configs:
        msg += f"* {current} (replaced with {default})\n"
    msg += "WARNING: Your changes in those files, if any, will be lost.\n"
    msg += "Proceed?"

    btn = KaliConfirmWindow(screen, msg, title=title)
    if btn != "ok":
        return

    # Do the actual job
    try:
        with suspend(screen, wait_enter=True):
            apply_shell_config_files(configs)
    except RuntimeError as e:
        KaliInfoWindow(screen, "error", str(e))
        return

    # One last notification
    say_reload_terminal(screen, "Your shell config files were reset!")


def do_shell_change_prompt(screen):
    title = "Command Prompt"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Get the current shell config
    config = get_current_shell_config()

    # Couldn't get the config?
    if config is None:
        msg = "Failed to parse shell configuration!\n"
        msg += "This is expected if your shell configuration files are"
        msg += " too old, and all you need to do is to reset it to the"
        msg += " latest default.\n"
        msg += "Do you want to reset your shell configuration files?"

        btn = KaliConfirmWindow(screen, msg, title=title)
        if btn == "ok":
            return "go-to-reset"
        else:
            return

    # Show the window with all the settings
    content = KaliShellPromptContent(config)
    w = KaliWindow(screen, title, content, buttons)
    btn = w.run_once()
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    msg = ""
    if "prompt" in changes:
        new = changes["prompt"]
        msg += f"Setting prompt to '{new}'.\n"
    if "newline" in changes:
        if changes["newline"] is True:
            msg += "Enabling new line before prompt.\n"
        else:
            msg += "Disabling new line before prompt.\n"
    msg += "Proceed?"

    btn = KaliConfirmWindow(screen, msg, title=title)
    if btn != "ok":
        return "retry"

    # Do the actual job
    try:
        with suspend(screen):
            apply_shell_config(changes)
    except RuntimeError as e:
        KaliInfoWindow(screen, "error", str(e))

    # One last notification
    say_reload_terminal(screen, "Your shell config files were updated!")


def do_virtual_setup(screen):
    virt = detect_virt()
    logger.debug("Virtualization: %s", virt)

    if virt is None:
        msg = "Couldn't detect virtualization type."
        KaliInfoWindow(screen, "info", msg)
        return

    pkg_to_install = None
    if virt.pkg and not is_pkg_installed(virt.pkg):
        pkg_to_install = virt.pkg

    scripts_to_install = virt.scripts

    if not pkg_to_install and not scripts_to_install:
        msg = f"Virtualization detected: {virt}.\n"
        msg += "Nothing to do."
        KaliInfoWindow(screen, "info", msg)
        return

    # Do the actual job
    try:
        msg = (
            "For more information about Virtual Machines, please refer to:\n"
            "https://www.kali.org/docs/virtualization/"
        )
        with suspend(screen, ending_message=msg, wait_enter=True):
            if pkg_to_install:
                install_pkgs(pkg_to_install)
            for script in scripts_to_install:
                install_program(script)
    except RuntimeError as e:
        KaliInfoWindow(screen, "error", str(e))


#### MAIN SCREENS


def do_shell_screen(screen):
    title = "Shell & Prompt"
    items = [
        ("Configure Prompt", "Configure the command prompt"),
        ("Default Login Shell", "Set the default login shell"),
        ("Reset Shell Config", "Reset the shell config files to their default"),
    ]
    buttons = [
        ("Select", "ok"),
        ("Back", "cancel"),
    ]
    ok_funcs = [
        do_shell_change_prompt,
        do_shell_change_default_login,
        do_shell_reset,
    ]

    while True:
        btn, line = KaliListboxWindow(screen, title, items, buttons)
        if btn != "ok":
            break
        func = ok_funcs[line]
        while True:
            ret = func(screen)
            if ret == "go-to-reset":
                func = do_shell_reset
                ret = func(screen)
            if ret != "retry":
                break


def do_virtual_screen(screen):
    title = "Virtualization"
    items = [
        ("Check and Configure", "Additional configuration for Virtual Machines"),
    ]
    buttons = [
        ("Select", "ok"),
        ("Back", "cancel"),
    ]
    ok_funcs = [
        do_virtual_setup,
    ]

    while True:
        btn, line = KaliListboxWindow(screen, title, items, buttons)
        if btn != "ok":
            break
        ok_funcs[line](screen)


def do_network_repo_screen(screen):
    title = "Network Repository"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Get the current apt config
    config = get_current_apt_sources_config()

    # Show the window with all the settings
    content = KaliNetworkRepoContent(config)
    w = KaliWindow(screen, title, content, buttons)
    btn = w.run_once()
    if btn != "ok":
        return

    # Get the settings that changed
    changes = content.getChanges()
    if changes == {}:
        return

    # Ask for confirmation
    msg = ""
    if "bleeding-edge" in changes:
        if changes["bleeding-edge"] is True:
            msg += "Enabling bleeding-edge repository.\n"
        else:
            msg += "Disabling bleeding-edge repository.\n"
    if "experimental" in changes:
        if changes["experimental"] is True:
            msg += "Enabling experimental repository.\n"
        else:
            msg += "Disabling experimental repository.\n"
    msg += "Proceed?"

    # Ask for confirmation
    btn = KaliConfirmWindow(screen, msg, title=title)
    if btn != "ok":
        return "retry"

    # Do the actual job
    try:
        msg = (
            "For more information about Kali repositories, please refer to:\n"
            "https://www.kali.org/docs/general-use/kali-linux-sources-list-repositories/"
        )
        with suspend(screen, ending_message=msg, wait_enter=True):
            apply_current_apt_sources_config(changes)
            apt_update()
    except RuntimeError as e:
        KaliInfoWindow(screen, "error", str(e))


def do_metapackages_screen(screen):
    title = "Metapackages"
    buttons = [
        ("Apply", "ok"),
        ("Back", "cancel"),
    ]

    # Check for kali-tools-* & kali-linux-*
    items = get_metapackage_items()

    # No Kali tools #1?
    if not items:
        apt_update()
        items = get_metapackage_items()

    # No Kali tools #2?
    if not items:
        msg = "No Kali Tools found. Are you really running Kali Linux? Or do you have an apt issue?"
        KaliInfoWindow(screen, "info", msg)
        return

    # Show the window with all the tools
    content = KaliSimpleCheckboxTree(items)
    w = KaliWindow(screen, title, content, buttons)
    btn = w.run_once()
    if btn != "ok":
        return

    # Sum up changes
    selection = content.getSelection()
    to_install = []
    to_remove = []
    for pkg, _, installed in items:
        package = "kali-tools-" + pkg
        package = package.replace("kali-tools-kali-linux-", "kali-linux-")
        if pkg in selection and not installed:
            to_install += [package]
        elif pkg not in selection and installed:
            to_remove += [package]

    # No change?
    if to_install == [] and to_remove == []:
        return

    # Ask for confirmation
    msg = ""
    if to_install:
        msg += "Packages to install: "
        msg += ", ".join(to_install)
        msg += "\n"
    if to_remove:
        msg += "Packages to remove: "
        msg += ", ".join(to_remove)
        msg += "\n"
    msg += "Proceed?"

    btn = KaliConfirmWindow(screen, msg, title=title)
    if btn != "ok":
        return "retry"

    # Do the actual job
    try:
        msg = (
            "For more information about Kali's metapackages, please refer to:\n"
            "https://www.kali.org/docs/general-use/metapackages/"
        )
        with suspend(screen, ending_message=msg, wait_enter=True):
            if to_remove:
                remove_pkgs(to_remove)
            if to_install:
                install_pkgs(to_install)
    except RuntimeError as e:
        KaliInfoWindow(screen, "error", str(e))


def do_main_screen(screen):
    title = "Main Menu"
    items = [
        ("Metapackages", "Install specific subsets of tools for particular needs"),
        ("Network Repositories", "Configure network repositories for APT sources"),
        ("Shell & Prompt", "Configure the shell and command prompt"),
        ("Virtualization", "Additional configurations for Virtual Machines"),
    ]
    buttons = [
        ("Select", "ok"),
        ("Quit", "cancel"),
    ]
    ok_funcs = [
        do_metapackages_screen,
        do_network_repo_screen,
        do_shell_screen,
        do_virtual_screen,
    ]

    while True:
        btn, line = KaliListboxWindow(screen, title, items, buttons)
        if btn != "ok":
            break
        func = ok_funcs[line]
        while True:
            ret = func(screen)
            if ret != "retry":
                break


def main():
    screen = SnackScreen()
    logger.debug(f"Screen dimensions: {screen.width} x {screen.height}")
    # Theming for checkbox and radiobox
    screen.setColor("CHECKBOX", "black", "lightgray")
    try:
        do_main_screen(screen)
    finally:
        screen.finish()
        # Print logs to stdout, if any
        if os.path.isfile(LOG_FILE):
            with open(LOG_FILE, "r") as f:
                print(f.read(), end="")
            os.remove(LOG_FILE)


if __name__ == "__main__":
    main()
