# Copyright 2021 Offensive Security
# SPDX-license-identifier: GPL-3.0-only

import errno
import getpass
import logging
import os
import re
import shutil
import subprocess
import tempfile
from pathlib import Path

logger = logging.getLogger(__name__)


def bold(text):
    bold = "\033[1m"
    reset = "\033[0m"
    return bold + text + reset


def say(text):
    print(bold(">>> " + text))
    logger.debug(text)


def say_install_packages(pkgs):
    if isinstance(pkgs, str):
        pkgs = [pkgs]
    if len(pkgs) == 0:
        text = "No package to install"
    elif len(pkgs) == 1:
        text = "Installing package: " + pkgs[0]
    else:
        text = "Installing packages: " + " ".join(pkgs)
    say(text)


def say_remove_packages(pkgs):
    if isinstance(pkgs, str):
        pkgs = [pkgs]
    if len(pkgs) == 0:
        text = "No package to remove"
    elif len(pkgs) == 1:
        text = "Removing package: " + pkgs[0]
    else:
        text = "Removing packages: " + " ".join(pkgs)
    say(text)


def say_install_program(pkgs):
    say(f"Installing program: {program}")


def say_enable_service(service):
    say(f"Enabling service: {service}")


def say_change_login_shell(shell):
    say(f"Setting login shell to '{shell}'")


def backup_filename(filename):
    """Return a backup filename"""
    return filename + ".kali-tweaks-orig"


def tmp_filename(filename):
    """Return a temporary filename"""
    # XXX Use proper tmp function
    return filename + ".tmp"


def _get_kali_tweaks_file(dir, filename):
    """Return the path to a kali-tweaks file"""
    if "RUN_FROM_SOURCE_TREE" in os.environ:
        f = os.path.join(os.getcwd(), dir, filename)
        if os.path.isfile(f):
            return f
    else:
        # path is hardcoded in setup.cfg, so it's hardcoded here as well
        f = os.path.join("/usr/lib/kali_tweaks", dir, filename)
        if os.path.isfile(f):
            return f
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), f)


def get_datafile(filename):
    """Return the path to a data file"""
    return _get_kali_tweaks_file("data", filename)


def get_helper(filename):
    """Return the path to a helper"""
    return _get_kali_tweaks_file("helpers", filename)


def _run(cmd, interactive=False):
    """If interactive is set, don't capture the output"""
    logger.debug("Running cmd: %s", cmd)
    res = subprocess.run(cmd, capture_output=(not interactive), shell=True, text=True)
    logger.debug("Return code: %d", res.returncode)
    # logger.debug("Stdout     :\n%s", res.stdout)
    # logger.debug("Stderr     :\n%s", res.stderr)
    return res


def run(cmd, **kwargs):
    assert not cmd.startswith("sudo")
    return _run(cmd, **kwargs)


def run_as_root(cmd, **kwargs):
    # no need to set interactive here, as sudo writes the prompt
    # straight on the terminal device, see sudo(8), option '-S'.
    rootcmd = "sudo " + cmd
    print("> $ " + rootcmd)
    return _run(rootcmd, **kwargs)


def apt_update():
    say("Updating APT cache")
    cmd = "apt-get update"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install package(s) '{pkgs}'")


def is_pkg_installed(pkg):
    res = run(f"dpkg -s {pkg} 2>/dev/null | grep -q 'ok installed'")
    return res.returncode == 0


def install_pkgs(pkgs, reinstall=False):
    say_install_packages(pkgs)
    cmd = get_helper("install-pkgs")
    if reinstall:
        cmd += " --reinstall"
    if isinstance(pkgs, list):
        pkgs = " ".join(pkgs)
    cmd += f" {pkgs}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install package(s) '{pkgs}'")


def remove_pkgs(pkgs):
    say_remove_packages(pkgs)
    cmd = "apt-get -y --purge remove"
    if isinstance(pkgs, list):
        pkgs = " ".join(pkgs)
    cmd += f" {pkgs}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't remove package(s) '{pkgs}'")


def is_program_installed(program):
    return os.path.isfile(program) and os.access(program, os.X_OK)


def install_program(program):
    say_install_program(program)
    dst = program
    src = get_datafile(os.path.basename(program))
    cmd = f"install -v -D -m0755 {src} {dst}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install program {src} -> {dst}")


def enable_service_now(service):
    say_enable_service(service)
    cmd = f"systemctl start {service}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't start service '{service}'")
    cmd = f"systemctl enable {service}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't enable service '{service}'")


#### SHELL HELPERS


def shell_from_shell_path(shell_path):
    """Get a shell name from a shell path, ie: /bin/zsh -> zsh"""
    for sh in ["zsh", "bash", "fish", "ksh", "sh", "tcsh"]:
        if f"bin/{sh}" in shell_path:
            return sh
    return None


def shell_path_from_shell(shell):
    """Get a shell path from a shell name, ie: zsh -> /usr/bin/zsh"""
    if shell == "bash":
        return "/usr/bin/bash"
    elif shell == "zsh":
        return "/usr/bin/zsh"
    else:
        return None


def get_configured_login_shell():
    """Get the login shell, as configured in /etc/passwd"""
    username = getpass.getuser()
    res = run(f"getent passwd {username}")
    if res.returncode != 0:
        logger.debug("getent passwd %s: %s", username, res.returncode)
        return None

    line = res.stdout.strip()
    _, _, shell_path = line.rpartition(":")
    # per passwd(5), the last field might be empty,
    # in such case it defaults to bin/sh
    if not shell_path:
        shell_path = "/bin/sh"
    logger.debug("Detected shell_path: %s (from /etc/passwd)", shell_path)

    shell = shell_from_shell_path(shell_path)
    logger.debug("Corresponding shell: %s", shell)
    return shell


def get_effective_login_shell():
    """Get the current login shell, as given by the 'SHELL' variable"""
    shell_path = os.getenv("SHELL")
    if shell_path is None:
        logger.debug("shell_path is None")
        return None
    logger.debug("Detected shell_path: %s (from SHELL)", shell_path)

    shell = shell_from_shell_path(shell_path)
    logger.debug("Corresponding shell: %s", shell)
    return shell


def change_login_shell(sh):
    """Change the login shell (requires logout+login to be effective)"""
    say_change_login_shell(sh)
    shell_path = shell_path_from_shell(sh)
    if shell_path is None:
        raise NotImplementedError(f"Can't set user shell to: {sh}")
    cmd = f"chsh -s {shell_path}"
    res = run(cmd, interactive=True)
    if res.returncode != 0:
        msg = f"Couldn't set login shell to '{sh}' ({shell_path})\n"
        msg += "Check console output for details."
        raise RuntimeError(msg)


def get_shell_config_files():
    """Return a list of tuples (default, current)"""
    configs = []
    homedir = str(Path.home())
    skeldir = "/etc/skel"
    for rc in [".bashrc", ".zshrc"]:
        default = os.path.join(skeldir, rc)
        current = os.path.join(homedir, rc)
        if not os.path.isfile(default):
            continue
        if not os.access(default, os.R_OK):
            continue
        configs += [(default, current)]
    return configs


def apply_shell_config_files(configs):
    """Take a list of tuples from get_shell_config_files()"""
    say("Resetting shell config files")
    try:
        for default, current in configs:
            if os.path.isfile(current):
                backup = backup_filename(current)
                print(f"> '{current}' -> '{backup}'")
                shutil.copyfile(current, backup)
            print(f"> '{default}' -> '{current}'")
            shutil.copyfile(default, current)
    except OSError as e:
        raise RuntimeError(str(e))


def _get_rc_file(filename):
    homedir = str(Path.home())
    path = os.path.join(homedir, filename)
    if not os.path.isfile(path):
        logger.debug("shell_rc is not a file")
        return None
    if not os.access(path, os.R_OK):
        logger.debug("no access to shell_rc")
        return None
    return path


def _get_shell_rc_file(shell):
    if shell == "bash":
        return _get_rc_file(".bashrc")
    elif shell == "zsh":
        return _get_rc_file(".zshrc")
    logger.debug("unknown rc name")
    return None


def _get_kali_config_block(data):
    """The whole config block as lines, including delimiters"""
    start_delimiter = "# START KALI CONFIG VARIABLES"
    stop_delimiter = "# STOP KALI CONFIG VARIABLES"
    res = []
    keep = False
    for line in data.splitlines():
        if line == start_delimiter:
            res = [line]
            keep = True
        elif line == stop_delimiter:
            res += [line]
            keep = False
            break
        elif keep:
            res += [line]

    if keep == True:
        # there was a start delimiter, but no stop delimiter
        return []

    return res


def _get_shell_config_from_file(filename):
    """Retrieve some config values from the shell config file"""
    # Get the configuration block
    with open(filename) as f:
        data = f.read()
    lines = _get_kali_config_block(data)
    if not lines:
        logger.debug("no lines in config")
        return None

    # Drop empty lines and comments
    lines = [l.strip(" \t") for l in lines]
    lines = [l for l in lines if l and not l.startswith("#")]

    # Make a dict out of that, drop anything unexpected.
    config = {}
    for l in lines:
        key, _, value = l.partition("=")
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        elif value.startswith("'") and value.endswith("'"):
            value = value[1:-1]
        if key == "NEWLINE_BEFORE_PROMPT":
            if value in ["yes", "no"]:
                config["newline"] = value == "yes"
        elif key == "PROMPT_ALTERNATIVE":
            if value in ["oneline", "twoline", "backtrack"]:
                config["prompt"] = value

    return config


def _apply_shell_config_to_file(filename, config):
    """Apply some config values to the shell config file"""
    # Get the configuration block
    with open(filename) as f:
        data = f.read()
    lines = _get_kali_config_block(data)
    if not lines:
        return

    # Keep a copy for later
    lines_orig = lines.copy()

    # Set some config (add it if missing)
    for key, value in config.items():
        if key == "newline":
            shkey = "NEWLINE_BEFORE_PROMPT"
            value = "yes" if value else "no"
        elif key == "prompt":
            shkey = "PROMPT_ALTERNATIVE"
        else:
            continue

        print(f"> Setting: {shkey}='{value}'")

        changed = False
        for idx, l in enumerate(lines):
            l = l.strip(" \t")
            if not l.startswith(shkey):
                continue
            lines[idx] = f"{shkey}='{value}'"
            changed = True
            break

        if changed:
            continue

        lines.append(f"{shkey}='{value}'")

    # Build a text block from lines
    config_block_orig = "\n".join(lines_orig)
    config_block = "\n".join(lines)

    logger.debug("config orig:\n%s", config_block_orig)
    logger.debug("config     :\n%s", config_block)

    # Return if no changes
    if config_block_orig == config_block:
        return

    # Do the actual replacement
    data = data.replace(config_block_orig, config_block)

    # Write to file
    tmp = tmp_filename(filename)
    with open(tmp, "w") as f:
        f.write(data)

    backup = backup_filename(filename)
    shutil.copyfile(filename, backup)
    shutil.move(tmp, filename)


def get_current_shell_config():
    """
    Retrieve the config for the current shell. Beware that config can
    be an *empty dict* (config file found, but no config variables in
    there), which is different from *None* (no config file found).
    """
    shell = get_effective_login_shell()
    if not shell:
        logger.debug("No shell detected")
        return None
    logger.debug("Detected shell: %s", shell)

    rc_file = _get_shell_rc_file(shell)
    if not rc_file:
        logger.debug("No shell_rc detected")
        return None
    logger.debug("Detected shell_rc: %s", rc_file)

    return _get_shell_config_from_file(rc_file)


def apply_shell_config(config):
    """Apply the config to the supported shells"""
    say("Configuring command prompt")
    for shell in ["bash", "zsh"]:
        rc_file = _get_shell_rc_file(shell)
        if not rc_file:
            continue
        print(f"> Updating: {rc_file}")
        _apply_shell_config_to_file(rc_file, config)


#### APT REPO HELPERS


def _get_listfile(suite):
    return f"/etc/apt/sources.list.d/kali-{suite}.list"


def _remove_listfile(suite):
    listfile = _get_listfile(suite)
    print(f"> Removed repository: {listfile}")
    cmd = f"rm -fv {listfile}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't remove file '{listfile}'")


def _install_listfile(suite):
    listfile = _get_listfile(suite)
    mirror = "http://http.kali.org/kali"
    components = "main contrib non-free"
    content = f"deb {mirror} kali-{suite} {components}"
    print(f"> New repository: {content}")
    print(f"> Installing to: '{listfile}'")
    f = tempfile.NamedTemporaryFile(mode="w", delete=False)
    f.write(content)
    f.close()
    cmd = f"install -m 0644 {f.name} {listfile}"
    res = run_as_root(cmd, interactive=True)
    if res.returncode != 0:
        raise RuntimeError(f"Couldn't install file to '{listfile}'")
    os.remove(f.name)


def get_current_apt_sources_config():
    """Retrieve APT sources configuration"""
    config = {}
    for suite in ["bleeding-edge", "experimental"]:
        listfile = _get_listfile(suite)
        if os.path.isfile(listfile):
            config[suite] = True
        else:
            config[suite] = False
    return config


def apply_current_apt_sources_config(config):
    """Apply changes to the to APT sources"""
    say("Configuring APT repositories")
    for suite in ["bleeding-edge", "experimental"]:
        if suite not in config:
            continue
        if config[suite] is True:
            _install_listfile(suite)
        else:
            _remove_listfile(suite)


#### VIRT HELPERS


class KvmVirt:
    def __init__(self):
        self.pkg = None
        self.scripts = []

    def __str__(self):
        return "KVM"


class QemuVirt:
    def __init__(self):
        self.pkg = None
        self.scripts = []

    def __str__(self):
        return "QEMU"


class VBoxVirt:
    def __init__(self):
        self.pkg = "virtualbox-guest-x11"
        self.scripts = []

    def __str__(self):
        return "VirtualBox"


class VMwareVirt:
    def __init__(self):
        self.pkg = "open-vm-tools-desktop"
        self.scripts = [
            "/usr/local/sbin/mount-share-folders",
            "/usr/local/sbin/restart-vm-tools",
        ]

    def __str__(self):
        return "VMWare"


def detect_virt():
    res = run("systemd-detect-virt")
    if res.returncode != 0:
        logger.debug("systemd-detect-virt: %s", res.returncode)
        return None

    vm = res.stdout.rstrip()
    logger.debug("Detected virt: %s", vm)

    if vm == "kvm":
        return KvmVirt()
    elif vm == "qemu":
        return QemuVirt()
    elif vm == "oracle":
        return VBoxVirt()
    elif vm in "vmware":
        return VMwareVirt()

    logger.debug("No VM detected")
    return None


#### DRY-RUN MODE, not documented, not maintained, don't rely on it


if "DRY_RUN" in os.environ and os.environ["DRY_RUN"] == "1":

    def install_program(program):
        say("Installing program: {program}")

    def install_pkgs(pkg, reinstall=False):
        say("Installing package: {pkg}")

    def remove_pkgs(pkg, reinstall=False):
        say("Removing package: {pkg}")

    def enable_service_now(service):
        say("Enabling service: {service}")
